/**
   \file
   \author Shin'ichiro Nakaoka
*/

#include "SceneProviderItemManager.h"
#include <cnoid/SceneProvider>
#include <cnoid/SceneGraph>
#include <cnoid/SceneView>
#include <cnoid/RootItem>
#include <cnoid/ItemTreeView>
#include <boost/bind.hpp>
#include <boost/tuple/tuple.hpp>
#include <map>

using namespace std;
using namespace boost;
using namespace cnoid;

namespace {
    const bool TRACE_FUNCTIONS = false;
}

namespace cnoid {

    class SceneProviderItemManagerImpl
    {
    public:
        SceneProviderItemManagerImpl(ExtensionManager* ext);
        ~SceneProviderItemManagerImpl();

        SceneView* sceneView;

        struct SceneInfo {
            SceneProvider* provider;
            SgNodePtr scene;
            bool isShown;
            signals::connection cSigDetachedFromRoot;
            signals::connection cSigCheckToggled;
            SceneInfo() : provider(0){
            }
            ~SceneInfo(){
                clear();
            }
            void clear() {
                cSigDetachedFromRoot.disconnect();
                cSigCheckToggled.disconnect();
                scene.reset();
            }
        };

        typedef map<ItemPtr, SceneInfo> SceneInfoMap;
        SceneInfoMap sceneInfoMap;

        signals::connection cSigItemAdded;

        void onItemAdded(Item* item);
        void removeScene(SceneInfoMap::iterator iter);
        void showScene(SceneInfo* info, bool show);
    };

    typedef SceneProviderItemManagerImpl Impl;
}


void SceneProviderItemManager::initialize(ExtensionManager* ext)
{
    static bool initialized = false;
    if(!initialized){
        ext->manage(new SceneProviderItemManager(ext));
        initialized = true;
    }
}

    
SceneProviderItemManager::SceneProviderItemManager(ExtensionManager* ext)
{
    impl = new Impl(ext);
}


Impl::SceneProviderItemManagerImpl(ExtensionManager* ext)
{
    sceneView = SceneView::instance();
    
    cSigItemAdded =
        RootItem::mainInstance()->sigItemAdded().connect(
            bind(&Impl::onItemAdded, this, _1));
}


SceneProviderItemManager::~SceneProviderItemManager()
{
    delete impl;
}


Impl::~SceneProviderItemManagerImpl()
{
    cSigItemAdded.disconnect();
}


void Impl::onItemAdded(Item* item)
{
    if(SceneProvider* provider = dynamic_cast<SceneProvider*>(item)){

        SceneInfoMap::iterator iter;
        bool inserted;
        tie(iter, inserted) = sceneInfoMap.insert(SceneInfoMap::value_type(item, SceneInfo()));
        SceneInfo* info = &iter->second;
        if(!inserted){
            info->clear();
        }
        info->provider = provider;
        info->isShown = false;
        
        info->cSigDetachedFromRoot = item->sigDetachedFromRoot().connect(
            bind(&Impl::removeScene, this, iter));
        
        ItemTreeView* itemTreeView = ItemTreeView::mainInstance();
        
        info->cSigCheckToggled = itemTreeView->sigCheckToggled(item).connect(
            bind(&Impl::showScene, this, info, _1));
        
        if(itemTreeView->isItemChecked(item)){
            showScene(info, true);
        }
    }
}


void Impl::removeScene(SceneInfoMap::iterator iter)
{
    SceneInfo* info = &iter->second;
    showScene(info, false);
    sceneInfoMap.erase(iter);
}

void Impl::showScene(SceneInfo* info, bool show)
{
    if(info->isShown && !show){
        if(info->scene){
            sceneView->removeEntity(info->scene);
        }
        info->isShown = false;
        
    } else if(!info->isShown && show){
        if(!info->scene){
            info->scene = info->provider->scene();
        }
        if(info->scene){
            sceneView->addEntity(info->scene);
            info->isShown = true;
        }
    }
}
