/**
   @author Shin'ichiro Nakaoka
*/

#include "SceneViewEditable.h"
#include "SceneView.h"

using namespace cnoid;

SceneViewEvent::SceneViewEvent()
{
    point_.setZero();
    x_ = 0.0;
    y_ = 0.0;
    key_ = 0;
    modifiers_ = 0;
    button_ = 0;
    wheelSteps_ = 0.0;
    viewMatrix_.setIdentity();
    projectionMatrix_.setIdentity();
    viewport_ << 0.0, 0.0, 1.0, 1.0;
}


SceneViewEvent::SceneViewEvent(const SceneViewEvent& org)
    : point_(org.point_)
{
    x_ = org.x_;
    y_ = org.y_;
    key_ = org.key_;
    modifiers_ = org.modifiers_;
    button_ = org.button_;
    wheelSteps_ = org.wheelSteps_;
    viewMatrix_ = org.viewMatrix_;
    projectionMatrix_ = org.projectionMatrix_;
    viewport_ = org.viewport_;
}


void SceneViewEvent::getViewport(int& out_x, int& out_y, int& out_width, int& out_height) const
{
    out_x = viewport_[0];
    out_y = viewport_[1];
    out_width = viewport_[2];
    out_height = viewport_[3];
}


bool SceneViewEvent::unproject(double x, double y, double z, Vector3& out_projected) const
{
    Vector4 p;
    p[0] = 2.0 * (x - viewport_[0]) / viewport_[2] - 1.0;
    p[1] = 2.0 * (y - viewport_[1]) / viewport_[3] - 1.0;
    p[2] = 2.0 * z - 1.0;
    p[3] = 1.0;
    
    const Vector4 projected = (projectionMatrix_ * viewMatrix_.matrix()).inverse() * p;

    if(projected[3] == 0.0){
        return false;
    }

    out_projected.x() = projected.x() / projected[3];
    out_projected.y() = projected.y() / projected[3];
    out_projected.z() = projected.z() / projected[3];

    return true;
}


void SceneViewEvent::updateIndicator(const std::string& text) const
{
    sceneView_->updateIndicator(text);
}


bool SceneViewEditable::onKeyPressEvent(const SceneViewEvent& event)
{
    return false;
}

bool SceneViewEditable::onKeyReleaseEvent(const SceneViewEvent& event)
{
    return false;
}

bool SceneViewEditable::onButtonPressEvent(const SceneViewEvent& event)
{
    return false;
}

bool SceneViewEditable::onButtonReleaseEvent(const SceneViewEvent& event)
{
    return false;
}

bool SceneViewEditable::onDoubleClickEvent(const SceneViewEvent& event)
{
    return false;
}

bool SceneViewEditable::onPointerMoveEvent(const SceneViewEvent& event)
{
    return false;
}

void SceneViewEditable::onPointerLeaveEvent(const SceneViewEvent& event)
{

}

bool SceneViewEditable::onScrollEvent(const SceneViewEvent& event)
{
    return false;
}

void SceneViewEditable::onContextMenuRequest(const SceneViewEvent& event, MenuManager& menuManager)
{

}

void SceneViewEditable::onSceneModeChanged(SceneMode mode)
{

}

bool SceneViewEditable::onUndoRequest()
{
    return false;
}

bool SceneViewEditable::onRedoRequest()
{
    return false;
}

void SceneViewEditable::setEditable(const bool on)
{
}

bool SceneViewEditable::isEditable()
{
    return false;
}


SceneProjector::~SceneProjector()
{
    
}


ScenePlaneProjector::ScenePlaneProjector()
{
    normal_ << 1.0, 0.0, 0.0;
    d_ = 0.0;
}


ScenePlaneProjector::ScenePlaneProjector(const Vector3& normal, const Vector3& point)
{
    setPlane(normal, point);
}


void ScenePlaneProjector::setPlane(const Vector3& normal, const Vector3& point)
{
    normal_ = normal;
    d_ = -normal.dot(point);
}


bool ScenePlaneProjector::project(const SceneViewEvent& event, Vector3& out_projected) const
{
    Vector3 nearPoint, farPoint;
    if(event.unproject(event.x(), event.y(), 0.0, nearPoint) &&
       event.unproject(event.x(), event.y(), 1.0, farPoint)){
        return calcPlaneLineIntersection(nearPoint, farPoint, out_projected);
    }
    return false;
}


bool ScenePlaneProjector::calcPlaneLineIntersection
(const Vector3& lineStart, const Vector3& lineEnd, Vector3& out_isect) const
{
    const Vector3& n = normal_;
    
    const double dx = lineEnd.x() - lineStart.x();
    const double dy = lineEnd.y() - lineStart.y();
    const double dz = lineEnd.z() - lineStart.z();

    const double denominator = (n.x() * dx + n.y() * dy + n.z() * dz);
    if(!denominator) {
        return false;
    }

    const double c = (n.x() * lineStart.x() + n.y() * lineStart.y() + n.z() * lineStart.z() + d_) / denominator;

    out_isect.x() = lineStart.x() - dx * c;
    out_isect.y() = lineStart.y() - dy * c;
    out_isect.z() = lineStart.z() - dz * c;

    return true;
}


SceneCylinderProjector::SceneCylinderProjector()
{
    center_ << 0.0, 0.0, 0.0;
    radius_ = 1.0;
    height_ = 1.0;
    rotation_.setIdentity();
}


SceneCylinderProjector::SceneCylinderProjector
(const Vector3& center, double radius, double height, const Quat& rotation)
{
    setCylinder(center, radius, height, rotation);
    rotation_ = rotation;
}


void SceneCylinderProjector::setCylinder
(const Vector3& center, double radius, double height, const Quat& rotation)
{
    center_ = center;
    radius_ = radius;
    height_ = height;
    rotation_ = rotation;
}


bool SceneCylinderProjector::project(const SceneViewEvent& event, Vector3& out_projected) const
{
    Vector3 nearPoint, farPoint;
    if(event.unproject(event.x(), event.y(), 0.0, nearPoint) &&
       event.unproject(event.x(), event.y(), 1.0, farPoint)){
        Vector3 isectBack;
        return calcCylinderLineIntersection(nearPoint, farPoint, out_projected, isectBack);
    }
    return false;
}


bool SceneCylinderProjector::calcUnitCylinderLineIntersection
(const Vector3& lineStart, const Vector3& lineEnd, Vector3& out_isectFront, Vector3& out_isectBack) const
{
    const Vector3 dir = (lineEnd - lineStart).normalized();
    const double a = dir[0] * dir[0] + dir[1] * dir[1];
    const double b = 2.0 * (lineStart[0] * dir[0] + lineStart[1] * dir[1]);
    const double c = lineStart[0] * lineStart[0] + lineStart[1] * lineStart[1] - 1.0;

    const double d = b * b - 4.0 * a * c;
    if(d < 0.0){
        return false;
    }

    const double dSqroot = sqrt(d);
    double t0, t1;
    if(b > 0.0){
        t0 = -(2.0 * c) / (dSqroot + b);
        t1 = -(dSqroot + b) / (2.0 * a);
    } else {
        t0 = (2.0 * c) / (dSqroot - b);
        t1 = (dSqroot - b) / (2.0 * a);
    }

    out_isectFront = lineStart + dir * t0;
    out_isectBack = lineStart + dir * t1;

    return true;
}


bool SceneCylinderProjector::calcCylinderLineIntersection
(const Vector3d& lineStart, const Vector3& lineEnd, Vector3& out_isectFront, Vector3& out_isectBack) const
{
    // Compute matrix transformation that takes the cylinder to a unit cylinder with Z-axis as it's axis and
    // (0,0,0) as it's center.
    const double rinv = 1.0 / radius_;
    const Affine3 toUnitCylInZ = rotation_.inverse() * Eigen::Scaling(rinv, rinv, rinv) * Translation3(-center_);
                               
    // Transform the lineStart and lineEnd into the unit cylinder space.
    const Vector3 unitCylLineStart = toUnitCylInZ * lineStart;
    const Vector3 unitCylLineEnd   = toUnitCylInZ * lineEnd;

    // Intersect line with unit cylinder.
    Vector3 unitCylIsectFront, unitCylIsectBack;
    if(!calcUnitCylinderLineIntersection(unitCylLineStart, unitCylLineEnd, unitCylIsectFront, unitCylIsectBack)){
        return false;
    }

    // Transform back from unit cylinder space.
    const Affine3 invToUnitCylInZ = toUnitCylInZ.inverse();
    out_isectFront = invToUnitCylInZ * unitCylIsectFront;
    out_isectBack = invToUnitCylInZ * unitCylIsectBack;    

    return true;
}
