/**
   @author Shin'ichiro Nakaoka
*/

#ifndef CNOID_BASE_SCENE_VIEW_EDITABLE__H_INCLUDED
#define CNOID_BASE_SCENE_VIEW_EDITABLE__H_INCLUDED

#include <cnoid/SceneGraph>
#include "exportdecl.h"

namespace cnoid {

    class SceneView;
    class MenuManager;
    
    class CNOID_EXPORT SceneViewEvent
    {
      public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW

        SceneViewEvent(const SceneViewEvent& org);
        
        inline const Vector3& point() const { return point_; }

        const SgNodePath& nodePath() const { return nodePath_; }

        // OpenGL viewport coordinate
        inline double x() const { return x_; }
        inline double y() const { return y_; }

        /**
           @return a Qt::Key value
        */
        inline int key() const { return key_; }

        /**
           @return a Qt::MouseButton value
        */
        inline int button() const { return button_; }

        /**
           @return a Qt::KeyboardModifiers value
        */
        inline int modifiers() const { return modifiers_; }
        
        inline double wheelSteps() const { return wheelSteps_; }

        const SgNodePath& cameraPath() const { return cameraPath_; }

        const Affine3& viewMatrix() const { return viewMatrix_; }
        const Matrix4& projectionMatrix() const { return projectionMatrix_; }
        const Array4i& viewport() const { return viewport_; }
        void getViewport(int& out_x, int& out_y, int& out_width, int& out_height) const;

        bool unproject(double x, double y, double z, Vector3& out_projected) const;
        Vector3 unproject(double x, double y, double z) const;
        
        void updateIndicator(const std::string& message) const;
        
      private:
        int key_;
        int button_;
        int modifiers_;
        Vector3 point_;
        double x_;
        double y_;
        double wheelSteps_;
        SgNodePath nodePath_;
        SgNodePath cameraPath_;
        Affine3 viewMatrix_;
        Matrix4 projectionMatrix_;
        Array4i viewport_;
        
        SceneView* sceneView_;

        SceneViewEvent();
        SceneViewEvent& operator=(const SceneViewEvent& org); // disabled

        friend class SceneViewImpl;
    };


    /**
       Interface class.
    */
    class CNOID_EXPORT SceneViewEditable
    {
    public:
        enum SceneMode { VIEW_MODE, EDIT_MODE };

        virtual bool onKeyPressEvent(const SceneViewEvent& event);
        virtual bool onKeyReleaseEvent(const SceneViewEvent& event);
        virtual bool onButtonPressEvent(const SceneViewEvent& event);
        virtual bool onButtonReleaseEvent(const SceneViewEvent& event);
        virtual bool onDoubleClickEvent(const SceneViewEvent& event);
        virtual bool onPointerMoveEvent(const SceneViewEvent& event);
        virtual void onPointerLeaveEvent(const SceneViewEvent& event);
        virtual bool onScrollEvent(const SceneViewEvent& event);
        virtual void onContextMenuRequest(const SceneViewEvent& event, MenuManager& menuManager);
        virtual void onSceneModeChanged(SceneMode mode);
        virtual bool onUndoRequest();
        virtual bool onRedoRequest();

        virtual void setEditable(const bool on);
        virtual bool isEditable();
    };


    class CNOID_EXPORT SceneProjector
    {
      public:
        virtual ~SceneProjector();
        virtual bool project(const SceneViewEvent& event, Vector3& out_projected) const = 0;
    };


    class CNOID_EXPORT ScenePlaneProjector : public SceneProjector
    {
      public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW

        ScenePlaneProjector();
        ScenePlaneProjector(const Vector3& normal, const Vector3& point);
        void setPlane(const Vector3& normal, const Vector3& point);

        const Vector3& normal() const { return normal_; }
        double d() const { return d_; }
        
        virtual bool project(const SceneViewEvent& event, Vector3& out_projected) const;

      private:
        Vector3 normal_;
        double d_;

        bool calcPlaneLineIntersection(
            const Vector3& lineStart, const Vector3& lineEnd, Vector3& out_isect) const;
    };


    class CNOID_EXPORT SceneCylinderProjector : public SceneProjector
    {
      public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW

        SceneCylinderProjector();
        SceneCylinderProjector(const Vector3& center, double radius, double height, const Quat& rotation);

        void setCylinder(const Vector3& center, double radius, double height, const Quat& rotation);
        
        virtual bool project(const SceneViewEvent& event, Vector3& out_projected) const;

      protected:
        Vector3 center_;
        double radius_;
        double height_;
        Quat rotation_;

        bool calcUnitCylinderLineIntersection(
            const Vector3& lineStart, const Vector3& lineEnd, Vector3& out_isectFront, Vector3& out_isectBack) const;
        bool calcCylinderLineIntersection(
            const Vector3d& lineStart, const Vector3& lineEnd, Vector3& out_isectFront, Vector3& out_isectBack) const;
    };
}

#endif
