/**
   \file
   \author Shin'ichiro Nakaoka
*/

#ifndef CNOID_BODY_BODY_H_INCLUDED
#define CNOID_BODY_BODY_H_INCLUDED

#include "LinkTraverse.h"
#include "Link.h"
#include "Device.h"
#include <cnoid/EigenTypes>
#include <cnoid/ValueTree>
#include <boost/any.hpp>
#include <map>
#include "exportdecl.h"

namespace cnoid {

    class Body;
    typedef boost::shared_ptr<Body> BodyPtr;

    struct BodyHandleEntity {
        Body* body;
    };
    struct BodyInterface;
    struct BodyCustomizerInterface;
    typedef void* BodyHandle;
    typedef void* BodyCustomizerHandle;

    class Device;

    class CNOID_EXPORT Body
    {
      public:
        static void addCustomizerDirectory(const std::string& path);
        static BodyInterface* bodyInterface();

        Body();
        Body(const Body& org);

        virtual BodyPtr clone() const;
        
        virtual Link* createLink(const Link* org = 0) const;

        virtual ~Body();

        const std::string& name() {
            return name_;
        }
        void setName(const std::string& name) {
            name_ = name;
        }
        const std::string& modelName() {
            return modelName_;
        }
        void setModelName(const std::string& name) {
            modelName_ = name;
        }
        
        void setRootLink(Link* link);

        /**
           This function must be called when the structure of the link tree is changed.
        */
        void updateLinkTree();

        void initializeState();

        /**
           The number of the links that work as a joint.
           Note that the acutal value is the maximum joint ID plus one.
           Thus there may be a case where the value does not correspond
           to the actual number of the joint-links.
           In other words, the value represents the size of the link sequence
           obtained by joint() function.
        */
        int numJoints() const {
            return jointIdToLinkArray.size();
        }

        /**
           This function returns a link that has a given joint ID.
           If there is no link that has a given joint ID,
           the function returns a dummy link object whose ID is minus one.
           The maximum id can be obtained by numJoints().
        */
        Link* joint(int id) const {
            return jointIdToLinkArray[id];
        }

        /**
           The number of all the links the body has.
           The value corresponds to the size of the sequence obtained by link() function.
        */
        int numLinks() const {
            return linkTraverse_.numLinks();
        }

        /**
           This function returns the link of a given index in the whole link sequence.
           The order of the sequence corresponds to a link-tree traverse from the root link.
           The size of the sequence can be obtained by numLinks().
        */
        Link* link(int index) const {
            return linkTraverse_.link(index);
        }

        /**
           LinkTraverse object that traverses all the links from the root link
        */
        const LinkTraverse& linkTraverse() const {
            return linkTraverse_;
        }

        /**
           This function returns a link object whose name of Joint node matches a given name.
           Null is returned when the body has no joint of the given name.
        */
        Link* link(const std::string& name) const {
            NameToLinkMap::const_iterator p = nameToLinkMap.find(name);
            return (p != nameToLinkMap.end()) ? p->second : 0;
        }

        /**
           The root link of the body
        */
        Link* rootLink() const {
            return rootLink_;
        }

        /**
           Example:
           const DeviceList<ForceSensor>& forceSensors = body->devices();
        */
        const DeviceList<>& devices() const {
            return devices_;
        }

        template<class DeviceType> DeviceType* device(const std::string& name) {
            DeviceType* device = 0;
            DeviceNameMap::iterator p = deviceNameMap.find(name);
            if(p != deviceNameMap.end()){
                device = dynamic_cast<DeviceType*>(p->second);
            }
            return device;
        }
        
        void addDevice(const DevicePtr& device);
        void initializeDeviceStates();
        
        /**
           This function returns true when the whole body is a static, fixed object like a floor.
        */
        bool isStaticModel() const {
            return isStaticModel_;
        }
        bool isFixedRootModel() const {
            return rootLink_->isFixedJoint();
        }

        double mass() const { return mass_; }

        const Vector3& calcCenterOfMass();
        const Vector3& centerOfMass() const { return centerOfMass_; }

        void calcTotalMomentum(Vector3& out_P, Vector3& out_L);

        void calcForwardKinematics(bool calcVelocity = false, bool calcAcceleration = false) {
                linkTraverse_.calcForwardKinematics(calcVelocity, calcAcceleration);
        }
        
        void clearExternalForces();

        bool hasVirtualJointForces() const;
        void setVirtualJointForces();

        /**
           This function must be called before the collision detection.
           It updates the positions and orientations of the models
           for detecting collisions between links.
        */
        void updateLinkColdetModelPositions();

        bool installCustomizer();
        bool installCustomizer(BodyCustomizerInterface* customizerInterface);

        enum ExtraJointType { EJ_PISTON, EJ_BALL };

        struct ExtraJoint {
            ExtraJointType type;
            Vector3 axis;
            Link* link[2];
            Vector3 point[2];
        };

        std::vector<ExtraJoint> extraJoints;

        const Mapping* info() const { return info_.get(); }
        Mapping* info() { return info_.get(); }
        void resetInfo(MappingPtr info);
        
        template<class T> T const * findCache(const std::string& name) const {
            CacheMap::const_iterator p = cacheMap.find(name);
            if(p != cacheMap.end()){
                T const * value = boost::any_cast<T>(&p->second);
                if(value){
                    return value;
                }
            }
            return 0;
        }

        template<class T> void setCache(const std::string& name, const T& cache) {
            cacheMap.insert(std::make_pair(name, boost::any(cache)));
        }
            
        void removeCache(const std::string& name) {
            cacheMap.erase(name);
        }

        BodyCustomizerHandle customizerHandle() const { return customizerHandle_; }
        BodyCustomizerInterface* customizerInterface() const { return customizerInterface_; }

        void resetDefaultPosition(const Position& T) { rootLink_->setOffsetPosition(T); }
        const Position& defaultPosition() const { return rootLink_->Tb(); }

#ifdef CNOID_BACKWARD_COMPATIBILITY
        BodyPtr duplicate() const { return clone(); }
        void initializeConfiguration() { initializeState(); }
        double calcTotalMass();
        Vector3 calcCM() { return calcCenterOfMass(); }
        double totalMass() const { return mass_; }
        const Vector3& lastCM() const { return centerOfMass_; }
        void setDefaultRootPosition(const Vector3& p, const Matrix3& R);
        void getDefaultRootPosition(Vector3& out_p, Matrix3& out_R);
#endif

      protected:
        void copy(const Body& org);

      private:
        Link* rootLink_;

        bool isStaticModel_;

        LinkTraverse linkTraverse_;

        typedef std::vector<Link*> LinkArray;
        LinkArray jointIdToLinkArray;

        typedef std::map<std::string, Link*> NameToLinkMap;
        NameToLinkMap nameToLinkMap;

        Vector3 centerOfMass_;
        double mass_;

        DeviceList<> devices_;

        typedef std::map<std::string, Device*> DeviceNameMap;
        DeviceNameMap deviceNameMap;
        
        // Members for customizer
        BodyCustomizerHandle customizerHandle_;
        BodyCustomizerInterface* customizerInterface_;
        BodyHandleEntity bodyHandleEntity;
        BodyHandle bodyHandle;

        MappingPtr info_;

        typedef std::map<std::string, boost::any> CacheMap;
        CacheMap cacheMap;

        Vector3 defaultRootPosition;
        Matrix3 defaultRootAttitude;

        std::string name_;
        std::string modelName_;
        
        void initialize();
        Link* cloneLinkTree(const Link* orgLink);
        Link* createEmptyJoint(int jointId);
        void setVirtualJointForcesSub();
    };
};

#endif
