/**
   \file
   \author Shin'ichiro Nakaoka
*/

#include "OSGSceneBodyManager.h"
#include "OSGSceneBody.h"
#include "LinkSelectionView.h"
#include <cnoid/RootItem>
#include <cnoid/ItemTreeView>
#include <cnoid/Archive>
#include <cnoid/MenuManager>
#include <osgDB/Registry>
#include <boost/signal.hpp>
#include <boost/bind.hpp>
#include <map>
#include <iostream>
#include <cassert>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

namespace {
    const bool TRACE_FUNCTIONS = false;

    OSGSceneBodyManager* instance_ = 0;
}

namespace cnoid {

    class SBMImpl : public boost::signals::trackable
    {
    public:
        SBMImpl(ExtensionManager& ext);

        struct SceneBodyInfo {
            BodyItem* bodyItem;
            OSGSceneBodyPtr sceneBody;
            bool isShown;
            signals::connection cSigDetachedFromRoot;
            signals::connection cSigCheckToggled;
            signals::connection cSigLinkSelectionChanged;
            ~SceneBodyInfo(){
                cSigDetachedFromRoot.disconnect();
                cSigCheckToggled.disconnect();
                cSigLinkSelectionChanged.disconnect();
            }
        };

        typedef list< boost::function<OSGSceneBody*(BodyItem*)> > FactoryList;
        FactoryList factories;
    
        typedef map<BodyItemPtr, SceneBodyInfo> SceneBodyInfoMap;
        SceneBodyInfoMap sceneBodyInfoMap;

        ItemTreeView* itemTreeView;

        Action* onlySelectedLinkCheck;

        void onItemAdded(Item* item);
        void onBodyItemDetached(BodyItem* item);
        void showBodyItem(SceneBodyInfo* info, bool show);
        void onOnlySelectedLinkToggled();
        void onLinkSelectionChanged(SceneBodyInfo* info);
        void saveSelectedSceneBodies();
        bool store(Archive& archive);
        void restore(const Archive& archive);
    };


    class FactoryHolderImpl : public OSGSceneBodyManager::FactoryHolder
    {
    public:
        FactoryHolderImpl(SBMImpl::FactoryList& factories, SBMImpl::FactoryList::iterator iter)
            : factories(factories), iter(iter) { }
        ~FactoryHolderImpl() { factories.erase(iter); }
    private:
        SBMImpl::FactoryList& factories;
        SBMImpl::FactoryList::iterator iter;
    };
}


OSGSceneBodyManager* OSGSceneBodyManager::instance()
{
    return instance_;
}


OSGSceneBodyManager::OSGSceneBodyManager(ExtensionManager& ext)
{
    assert(!instance_);
    impl = new SBMImpl(ext);
    instance_ = this;
}


SBMImpl::SBMImpl(ExtensionManager& ext)
{
    RootItem::mainInstance()->sigItemAdded().connect(bind(&SBMImpl::onItemAdded, this, _1));

    itemTreeView = ItemTreeView::mainInstance();

    ext.connectProjectArchiver(
        "OSGSceneBodyManager", bind(&SBMImpl::store, this, _1), bind(&SBMImpl::restore, this, _1));

    onlySelectedLinkCheck = ext.menuManager().setPath("/Options/Scene View (OSG)").addCheckItem(_("Show only selected links"));
    onlySelectedLinkCheck->setChecked(false);
    onlySelectedLinkCheck->sigToggled().connect(bind(&SBMImpl::onOnlySelectedLinkToggled, this));

    ext.menuManager().addItem(_("Save scenes bodies"))->sigTriggered().connect(
        bind(&SBMImpl::saveSelectedSceneBodies, this));
}


OSGSceneBodyManager::~OSGSceneBodyManager()
{
    delete impl;
}


OSGSceneBody* OSGSceneBodyManager::sceneBody(BodyItemPtr bodyItem)
{
    SBMImpl::SceneBodyInfoMap::iterator p = impl->sceneBodyInfoMap.find(bodyItem);
    if(p != impl->sceneBodyInfoMap.end()){
        return p->second.sceneBody.get();
    }
    return 0;
}
    

/**
   @param factory A factory function object that creates a scene body instance of a customized sub class.
   In order to avoid the collision between the multiple factories,
   it is desirable that the factory function checks the actual type of a BodyItem instance and
   a customized scene body is only returned if the type matches. Example code is as follows:
   \code
   OSGSceneBody* factory(BodyItem* bodyItem)
   {
       BodyItemEx* ex = dynamic_cast<BodyItemEx*>(bodyItem);
       if(ex){
           return new OSGSceneBodyEx(ex);
       }
       return 0;
   }
   \endcode
   
   @return A holder object of the registered factory. The destructor of the holder object unregisters the factory.
   This object is usually passed to 'ExtensionManager::manage()' function so that the factory can be
   automatically removed when the plugin is finalized.
*/
OSGSceneBodyManager::FactoryHolder* OSGSceneBodyManager::addSceneBodyFactory(boost::function<OSGSceneBody*(BodyItem*)> factory)
{
    impl->factories.push_front(factory);
    return new FactoryHolderImpl(impl->factories, --impl->factories.end());
}


void SBMImpl::onItemAdded(Item* item)
{
    if(TRACE_FUNCTIONS){
        cout << "SBMImpl::onItemAdded(" << item << ")" << endl;
    }
    
    if(BodyItem* bodyItem = dynamic_cast<BodyItem*>(item)){

        SceneBodyInfo* info = &sceneBodyInfoMap[bodyItem];

        info->bodyItem = bodyItem;
        if(!info->sceneBody){
            for(FactoryList::iterator p = factories.begin(); p != factories.end(); ++p){
                info->sceneBody = (*p)(bodyItem);
                if(info->sceneBody.valid()){
                    break;
                }
            }
        }
        if(!info->sceneBody){
            info->sceneBody = new OSGSceneBody(bodyItem);
        }

        info->isShown = false;

        info->cSigDetachedFromRoot = bodyItem->sigDetachedFromRoot().connect(
            bind(&SBMImpl::onBodyItemDetached, this, bodyItem));

        info->cSigCheckToggled = itemTreeView->sigCheckToggled(bodyItem).connect(
            bind(&SBMImpl::showBodyItem, this, info, _1));

        if(itemTreeView->isItemChecked(bodyItem)){
            showBodyItem(info, true);
        }
    }
}


void SBMImpl::onBodyItemDetached(BodyItem* item)
{
    if(TRACE_FUNCTIONS){
        cout << "SBMImpl::onBodyItemDetached(" << item << ")" << endl;
    }

    SceneBodyInfoMap::iterator p = sceneBodyInfoMap.find(item);
    if(p != sceneBodyInfoMap.end()){
        SceneBodyInfo* info = &p->second;
        showBodyItem(info, false);
        sceneBodyInfoMap.erase(p);
    }
}


void SBMImpl::showBodyItem(SceneBodyInfo* info, bool show)
{
    if(TRACE_FUNCTIONS){
        cout << "SBMImpl::showBodyItem()" << endl;
    }

    if(info->isShown && !show){
        info->cSigLinkSelectionChanged.disconnect();
        info->sceneBody->hide();
        info->isShown = false;
        
    } else if(!info->isShown && show){
        info->sceneBody->show();
        info->isShown = true;

        info->cSigLinkSelectionChanged =
            LinkSelectionView::mainInstance()->sigSelectionChanged(info->bodyItem).connect(
                bind(&SBMImpl::onLinkSelectionChanged, this, info));
        onLinkSelectionChanged(info);
    }
}


void SBMImpl::onOnlySelectedLinkToggled()
{
    SceneBodyInfoMap::iterator p;
    for(p = sceneBodyInfoMap.begin(); p != sceneBodyInfoMap.end(); ++p){
        if(onlySelectedLinkCheck->isChecked()){
            onLinkSelectionChanged(&p->second);
        } else {
            dynamic_bitset<> visibilities(p->first->body()->numLinks(), true);
            p->second.sceneBody->setLinkVisibilities(visibilities);
        }
    }
}


void SBMImpl::saveSelectedSceneBodies()
{
    ItemList<BodyItem> bodyItems = itemTreeView->selectedItems<BodyItem>();
    for(size_t i=0; i < bodyItems.size(); ++i){
        BodyItem* bodyItem = bodyItems.get(i);
        SceneBodyInfoMap::iterator p = sceneBodyInfoMap.find(bodyItem);
        if(p != sceneBodyInfoMap.end()){

            osg::ref_ptr<osg::Group> group = new osg::Group;
            OSGSceneBodyPtr scene = p->second.sceneBody;
            int n = scene->getNumChildren();
            for(int i=0; i < n; ++i){
                group->addChild(scene->getChild(i));
            }
            osgDB::ReaderWriter::WriteResult result =
                osgDB::Registry::instance()->writeNode(*group, bodyItem->name() + ".osgt", 0);
            if(!result.message().empty()){
                mvout() << result.message() << endl;
            }
        }
    }
}


void SBMImpl::onLinkSelectionChanged(SceneBodyInfo* info)
{
    if(onlySelectedLinkCheck->isChecked()){
        info->sceneBody->setLinkVisibilities(
            LinkSelectionView::mainInstance()->getLinkSelection(info->bodyItem));
    }
}


bool SBMImpl::store(Archive& archive)
{
    ListingPtr states = new Listing();
    
    SceneBodyInfoMap::iterator p;
    for(p = sceneBodyInfoMap.begin(); p != sceneBodyInfoMap.end(); ++p){
        BodyItem* bodyItem = p->first.get();
        int id = archive.getItemId(bodyItem);
        if(id >= 0){
            OSGSceneBodyPtr sceneBody = p->second.sceneBody;
            MappingPtr state = new Mapping();
            state->write("bodyItem", id);
            state->write("editable", sceneBody->isEditable());
            state->write("centerOfMass", sceneBody->isCenterOfMassVisible());
            state->write("zmp", sceneBody->isZmpVisible());
            states->append(state);
        }
    }

    if(!states->empty()){
        archive.insert("sceneBodies", states);
        return true;
    }

    return false;
}


void SBMImpl::restore(const Archive& archive)
{
    Listing& states = *archive["sceneBodies"].toListing();
    for(int i=0; i < states.size(); ++i){
        Mapping* state = states[i].toMapping();
        BodyItem* bodyItem = archive.findItem<BodyItem>(state->get("bodyItem", -1));
        if(bodyItem){
            SceneBodyInfoMap::iterator p = sceneBodyInfoMap.find(bodyItem);
            if(p != sceneBodyInfoMap.end()){
                OSGSceneBodyPtr sceneBody = p->second.sceneBody;
                sceneBody->setEditable(state->get("editable", sceneBody->isEditable()));
                sceneBody->showCenterOfMass(state->get("centerOfMass", sceneBody->isCenterOfMassVisible()));
                sceneBody->showZmp(state->get("zmp", sceneBody->isZmpVisible()));
            }
        }
    }
}
