/**
   \file
   \author Shizuko Hattori
*/

#include "VirtualRobotPortHandler.h"
#include "BodyRTCItem.h"
#include <cnoid/EigenUtil>
#include <cnoid/DyBody>

using namespace std;
using namespace RTC;
using namespace cnoid;


PortHandler::~PortHandler()
{

}


SensorStateOutPortHandler::SensorStateOutPortHandler(PortInfo& info) : 
  OutPortHandler(info), 
  outPort(info.portName.c_str(), values)
{
  dataTypeId = info.dataTypeId;
  stepTime = info.stepTime;
}

void SensorStateOutPortHandler::inputDataFromSimulator(BodyRTCItem* bodyRTC)
{
    const BodyPtr& body = bodyRTC->body();
    switch(dataTypeId) {
    case JOINT_VALUE:
    {
        const int n = body->numJoints();
        values.data.length(n);
        for(int i=0; i < n; i++){
            values.data[i] = body->joint(i)->q();
        }
        break;
    }
    case JOINT_VELOCITY:
    {
        const int n =body->numJoints();
        values.data.length(n);
        for(int i=0; i < n; i++){
            values.data[i] = body->joint(i)->dq();
        }
        break;
    }
    case JOINT_TORQUE:
    {
        const int n = body->numJoints();
        values.data.length(n);
        for(int i=0; i < n; i++){
            values.data[i] = body->joint(i)->u();
        }
        break;
    }
    case FORCE_SENSOR:
    {
        const DeviceList<ForceSensor>& forceSensors = bodyRTC->forceSensors();
        const int n = forceSensors.size();
        values.data.length(6 * n);
        for(int i=0; i < n; i++){
            Eigen::Map<Vector6>(&values.data[i*6]) = forceSensors[i]->F();
        }
    }
    break;
    case RATE_GYRO_SENSOR:
    {
        const DeviceList<RateGyroSensor>& gyroSensors = bodyRTC->gyroSensors();
        const int n = gyroSensors.size();
        values.data.length(3 * n);
        for(int i=0; i < n; i++){
            Eigen::Map<Vector3>(&values.data[i*3]) = gyroSensors[i]->w();
        }
    }
    break;
    case ACCELERATION_SENSOR:
    {
        const DeviceList<AccelSensor>& accelSensors = bodyRTC->accelSensors();
        const int n = accelSensors.size();
        values.data.length(3 * n);
        for(int i=0; i < n; i++){
            Eigen::Map<Vector3>(&values.data[i*3]) = accelSensors[i]->dv();
        }
    }
    break;
    default:
        break;
    }
    setTime(values, bodyRTC->controlTime());
}

void SensorStateOutPortHandler::writeDataToPort()
{
  outPort.write();
}


LinkDataOutPortHandler::LinkDataOutPortHandler(PortInfo& info) :
  OutPortHandler(info),
  outPort(info.portName.c_str(), value),
  linkName(info.dataOwnerName)
{
  linkDataType = info.dataTypeId;
  stepTime = info.stepTime;
}


void LinkDataOutPortHandler::inputDataFromSimulator(BodyRTCItem* bodyRTC)
{
    const BodyPtr& body = bodyRTC->body();
    size_t n = linkName.size();
    switch(linkDataType) {
    case JOINT_VALUE:
        value.data.length(n);
        for(size_t i=0; i<n; i++){
            value.data[i] = body->link(linkName[i])->q();
        }
        break;
    case JOINT_VELOCITY:
        value.data.length(n);
        for(size_t i=0; i<n; i++){
            value.data[i] = body->link(linkName[i])->dq();
        }
        break;
    case JOINT_ACCELERATION:
        value.data.length(n);
        for(size_t i=0; i<n; i++){
            value.data[i] = body->link(linkName[i])->ddq();
        }
        break;
    case JOINT_TORQUE:
        value.data.length(n);
        for(size_t i=0; i<n; i++){
            value.data[i] = body->link(linkName[i])->u();
        }
        break;
    case ABS_TRANSFORM:
        value.data.length(12*n);
        for(size_t i=0, j=0; i<n; i++){
            Link* link = body->link(linkName[i]);
            value.data[j++] = link->p()(0);
            value.data[j++] = link->p()(1);
            value.data[j++] = link->p()(2);
            const Matrix3 m33 = link->attitude();
            value.data[j++] = m33(0, 0);
            value.data[j++] = m33(0, 1);
            value.data[j++] = m33(0, 2);
            value.data[j++] = m33(1, 0);
            value.data[j++] = m33(1, 1);
            value.data[j++] = m33(1, 2);
            value.data[j++] = m33(2, 0);
            value.data[j++] = m33(2, 1);
            value.data[j++] = m33(2, 2);
        }
        break;
    case ABS_VELOCITY:
    {
        value.data.length(6*n);
        for(size_t i=0, j=0; i<n; i++){
            Link* link = body->link(linkName[i]);
            value.data[j++] = link->v()(0);
            value.data[j++] = link->v()(1);
            value.data[j++] = link->v()(2);
            value.data[j++] = link->w()(0);
            value.data[j++] = link->w()(1);
            value.data[j++] = link->w()(2);
        }
    }
    break;
    case EXTERNAL_FORCE:
    {
        value.data.length(6*n);
        for(size_t i=0, j=0; i<n; i++){
            Link* link = body->link(linkName[i]);
            value.data[j++] = link->f_ext()(0);
            value.data[j++] = link->f_ext()(1);
            value.data[j++] = link->f_ext()(2);
            value.data[j++] = link->tau_ext()(0);
            value.data[j++] = link->tau_ext()(1);
            value.data[j++] = link->tau_ext()(2);
        }
    }
    break;
    case CONSTRAINT_FORCE:   // linkName size must be one.
    {
        DyLink* link = dynamic_cast<DyLink*>(body->link(linkName[0]));
        if(link){
            const vector<DyLink::ConstraintForce>& constraintForces = link->constraintForces();
            const size_t n = constraintForces.size();
            value.data.length(6*n);
            for(size_t i=0, j=0; i<n; i++){
                value.data[j++] = constraintForces[i].point(0);
                value.data[j++] = constraintForces[i].point(1);
                value.data[j++] = constraintForces[i].point(2);
                value.data[j++] = constraintForces[i].force(0);
                value.data[j++] = constraintForces[i].force(1);
                value.data[j++] = constraintForces[i].force(2);
            }
        }
    }
    break;
    default:
        break;
    }
    setTime(value, bodyRTC->controlTime());
}

void LinkDataOutPortHandler::writeDataToPort()
{
  outPort.write();
}

AbsTransformOutPortHandler::AbsTransformOutPortHandler(PortInfo& info) :
  OutPortHandler(info),
  outPort(info.portName.c_str(), value),
  linkName(info.dataOwnerName)
{
  linkDataType = info.dataTypeId;
  stepTime = info.stepTime;
}


void AbsTransformOutPortHandler::inputDataFromSimulator(BodyRTCItem* bodyRTC)
{
    Link* link = bodyRTC->body()->link(linkName[0]);
    value.data.position.x = link->p().x();
    value.data.position.y = link->p().y();
    value.data.position.z = link->p().z();
    Matrix3 R = link->attitude();
    const Vector3 rpy = rpyFromRot(R);
    value.data.orientation.r = rpy[0];
    value.data.orientation.p = rpy[1];
    value.data.orientation.y = rpy[2];
    setTime(value, bodyRTC->controlTime());
}


void AbsTransformOutPortHandler::writeDataToPort()
{
  outPort.write();
}


SensorDataOutPortHandler::SensorDataOutPortHandler(PortInfo& info) :
  OutPortHandler(info),
  outPort(info.portName.c_str(), value),
  sensorName(info.dataOwnerName)
{
    stepTime = info.stepTime;
}


void SensorDataOutPortHandler::inputDataFromSimulator(BodyRTCItem* bodyRTC)
{
    const BodyPtr& body = bodyRTC->body();
    if(!sensorName.empty()){
        if(Sensor* sensor = body->device<Sensor>(sensorName[0])){
            const int dataSize = sensor->stateSize();
            value.data.length(dataSize);
            if(dataSize > 0){
                for(size_t i=0; i < sensorName.size(); ++i){
                    if(Sensor* sensor = body->device<Sensor>(sensorName[i])){
                        sensor->writeState(&value.data[i * dataSize]);
                    }
                }
            }
        }
    }
    setTime(value, bodyRTC->controlTime());
}

void SensorDataOutPortHandler::writeDataToPort()
{
  outPort.write();
}


GyroSensorOutPortHandler::GyroSensorOutPortHandler(PortInfo& info) :
    OutPortHandler(info),
    outPort(info.portName.c_str(), value),
    sensorName(info.dataOwnerName)
{
    stepTime = info.stepTime;
}


void GyroSensorOutPortHandler::inputDataFromSimulator(BodyRTCItem* bodyRTC)
{
    if(!sensorName.empty()){
        if(RateGyroSensor* gyroSensor =bodyRTC->body()->device<RateGyroSensor>(sensorName[0])){
            value.data.avx = gyroSensor->w().x();
            value.data.avy = gyroSensor->w().y();
            value.data.avz = gyroSensor->w().z();
            setTime(value, bodyRTC->controlTime());
        }
    }
}


void GyroSensorOutPortHandler::writeDataToPort()
{
    outPort.write();
}


AccelerationSensorOutPortHandler::AccelerationSensorOutPortHandler(PortInfo& info) :
    OutPortHandler(info),
    outPort(info.portName.c_str(), value),
    sensorName(info.dataOwnerName)
{
    stepTime = info.stepTime;
}


void AccelerationSensorOutPortHandler::inputDataFromSimulator(BodyRTCItem* bodyRTC)
{
    if(!sensorName.empty()){
        if(AccelSensor* accelSensor = bodyRTC->body()->device<AccelSensor>(sensorName[0])){
            value.data.ax = accelSensor->dv().x();
            value.data.ay = accelSensor->dv().y();
            value.data.az = accelSensor->dv().z();
            setTime(value, bodyRTC->controlTime());
        }
    }
}


void AccelerationSensorOutPortHandler::writeDataToPort()
{
    outPort.write();
}


ColorImageOutPortHandler::ColorImageOutPortHandler(PortInfo& info) :
    OutPortHandler(info),
    outPort(info.portName.c_str(), image),
    cameraId(info.dataOwnerId)
{
    stepTime = info.stepTime;
}


void ColorImageOutPortHandler::inputDataFromSimulator(BodyRTCItem* bodyRTC)
{
    //ImageData_var imageInput = controller->getCameraImageFromSimulator(cameraId);
    //image.data = imageInput->longData;
    setTime(image, bodyRTC->controlTime());
}


void ColorImageOutPortHandler::writeDataToPort()
{
    outPort.write();
}


GrayScaleImageOutPortHandler::GrayScaleImageOutPortHandler(PortInfo& info) : 
    OutPortHandler(info),
    outPort(info.portName.c_str(), image),
    cameraId(info.dataOwnerId)
{
    stepTime = info.stepTime;
}


void GrayScaleImageOutPortHandler::inputDataFromSimulator(BodyRTCItem* bodyRTC)
{
    //ImageData_var imageInput = controller->getCameraImageFromSimulator(cameraId);
    //image.data = imageInput->octetData;
    setTime(image, bodyRTC->controlTime());
}


void GrayScaleImageOutPortHandler::writeDataToPort()
{
    outPort.write();
}


DepthImageOutPortHandler::DepthImageOutPortHandler(PortInfo& info) :
    OutPortHandler(info),
    outPort(info.portName.c_str(), image),
    cameraId(info.dataOwnerId)
{
    stepTime = info.stepTime;
}


void DepthImageOutPortHandler::inputDataFromSimulator(BodyRTCItem* bodyRTC)
{
    //ImageData_var imageInput = controller->getCameraImageFromSimulator(cameraId);
    //image.data = imageInput->floatData;
    setTime(image, bodyRTC->controlTime());
}


void DepthImageOutPortHandler::writeDataToPort()
{
    outPort.write();
}

LightOnOutPortHandler::LightOnOutPortHandler(PortInfo& info) :
  OutPortHandler(info),
  outPort(info.portName.c_str(), value),
  lightName(info.dataOwnerName)
{
    stepTime = info.stepTime;
}


void LightOnOutPortHandler::inputDataFromSimulator(BodyRTCItem* bodyRTC)
{
    const BodyPtr& body = bodyRTC->body();
    value.data.length(lightName.size());
    size_t i=0;
    for(vector<string>::iterator it = lightName.begin(); it != lightName.end(); it++){
        Light* light = body->device<Light>(*it);
        value.data[i++] = light->on(); 
    }
    
    setTime(value, bodyRTC->controlTime());
}

void LightOnOutPortHandler::writeDataToPort()
{
  outPort.write();
}

JointDataSeqInPortHandler::JointDataSeqInPortHandler(PortInfo& info) :
    InPortHandler(info),
    inPort(info.portName.c_str(), values)
{
    linkDataType = info.dataTypeId;
}


void JointDataSeqInPortHandler::outputDataToSimulator(const BodyPtr& body)
{
    size_t n =body->numJoints();
    if(values.data.length()!=n)
        return;
    switch(linkDataType) {
    case JOINT_VALUE:
        for(size_t i=0; i<n; i++)
            body->joint(i)->q() = values.data[i];
        break;
    case JOINT_VELOCITY:
        for(size_t i=0; i<n; i++)
            body->joint(i)->dq() = values.data[i];
        break;
    case JOINT_ACCELERATION:
        for(size_t i=0; i<n; i++)
            body->joint(i)->ddq() = values.data[i];
        break;
    case JOINT_TORQUE:
        for(size_t i=0; i<n; i++)
            body->joint(i)->u() = values.data[i];
        break;
    default :
        break;
    }
}


void JointDataSeqInPortHandler::readDataFromPort()
{
    if( inPort.isNew() == false ){
        values.data.length(0);
    }else
	inPort.read();
}

LinkDataInPortHandler::LinkDataInPortHandler(PortInfo& info) :
    InPortHandler(info),
    inPort(info.portName.c_str(), values),
    linkName(info.dataOwnerName)
{
    linkDataType = info.dataTypeId;
}

void LinkDataInPortHandler::outputDataToSimulator(const BodyPtr& body)
{
    size_t n=linkName.size();
    if(values.data.length()!=n)
        return;
    switch(linkDataType) {
    case JOINT_VALUE:
        for(size_t i=0; i<n; i++)
            body->link(linkName[i])->q() = values.data[i];
        break;
    case JOINT_VELOCITY:
        for(size_t i=0; i<n; i++)
            body->link(linkName[i])->dq() = values.data[i];
        break;
    case JOINT_ACCELERATION:
        for(size_t i=0; i<n; i++)
            body->link(linkName[i])->ddq() = values.data[i];
        break;
    case JOINT_TORQUE:
        for(size_t i=0; i<n; i++)
            body->link(linkName[i])->u() = values.data[i];
        break;
    default :
        break;
    }
}

void LinkDataInPortHandler::readDataFromPort()
{
    if( inPort.isNew() == false ){
	values.data.length(0);
    }else
	inPort.read();
}

AbsTransformInPortHandler::AbsTransformInPortHandler(PortInfo& info) :
    InPortHandler(info),
    inPort(info.portName.c_str(), values),
    linkName(info.dataOwnerName)
{
    linkDataType = info.dataTypeId;
}


void AbsTransformInPortHandler::outputDataToSimulator(const BodyPtr& body)
{
    Link* link = body->link(linkName[0]);
    link->p().x() = values.data.position.x;
    link->p().y() = values.data.position.y;
    link->p().z() = values.data.position.z;
    Matrix3 R = rotFromRpy(values.data.orientation.r,
                           values.data.orientation.p,
                           values.data.orientation.y);
    link->setAttitude(R);
    
}

void AbsTransformInPortHandler::readDataFromPort()
{
    if( inPort.isNew() == false ){
        return;
    }else
	inPort.read();
}

LightOnInPortHandler::LightOnInPortHandler(PortInfo& info) :
    InPortHandler(info),
    inPort(info.portName.c_str(), values),
    lightName(info.dataOwnerName)
{
}

void LightOnInPortHandler::outputDataToSimulator(const BodyPtr& body)
{
    for(size_t i=0; i<values.data.length(); i++){
        Light* light = body->device<Light>(lightName[i]);
        light->on(values.data[i]);
        light->notifyStateChange();
    }
}

void LightOnInPortHandler::readDataFromPort()
{
    if( inPort.isNew() == false ){
        values.data.length(0);
    }else
	inPort.read();
}
